"""

"""

from . import util

__all__ = [
    "CacheControl",
    "Connection",
    "Date",
    "Pragma",
    "Trailer",
    "TransferEncoding",
    "Upgrade",
    "Via",
    "Warning",
]


class General(util.Header):

    """"""


class CacheControl(General):

    """"""


class Connection(General):

    """"""


class Date(General):

    """"""


class Pragma(General):

    """"""


class Trailer(General):

    """"""


class TransferEncoding(General):

    """"""


class Upgrade(General):

    """"""


class Via(General):

    """"""


class Warning(General):

    """"""
