/*

  base03   #002b36  * dark
  base02   #073642  * dark
  
  base01   #586e75  * both
  
  base00   #657b83  * light
  base0    #839496  * dark
  
  base1    #93a1a1  * both
  
  base2    #eee8d5  * light
  base3    #fdf6e3  * light
  
  magenta  #d33682    green     #859900
  violet   #6c71c4    yellow    #b58900
  blue     #268bd2    orange    #cb4b16
  cyan     #2aa198    red       #dc322f

*/

.dark .highlight {
  background-color: #002b36; }

.dark .highlight .code {
  color: #839496; }
.light .highlight .code {
  color: #657b83; }
.highlight .code {
  width: 100%; }

.dark .highlighttable .linenos a {
    color: #586e75; }
.light .highlighttable .linenos a {
    color: #93a1a1; }
.linenos {
  padding: 0 1em;
  text-align: right; }
.linenos a {
  text-decoration: none; }

.dark .highlighttable .focus {
    background-color: #073642;
    /* border-right: .25em solid #2aa198; */
    right: .25em; }

/* .dark .highlighttable .covered {
    background-color: #073642;
    border-right: .25em solid #859900; }
.light .highlighttable .covered {
    color: ; } */
.dark .highlighttable .uncovered {
    background-color: #073642;
    border-right: .25em solid #d33682; }
/* .light .highlighttable .uncovered {
    color: ; } */
.dark .highlighttable .failed {
    background-color: #073642;
    border-right: .25em solid #dc322f; }
/* .light .highlighttable .failed {
    color: ; } */

.hll { background-color: #ffffcc }
.c { color: #586e75; font-style: italic } /* Comment */
.err { color: #dc322f } /* Error */
.g { color: #839496 } /* Generic */
.k { color: #859900 } /* Keyword */
.l { color: #839496 } /* Literal */
.dark .n { color: #839496; } /* Name */
.light .n { color: #657b83; } /* Name */
.o { color: #839496 } /* Operator */
.x { color: #839496 } /* Other */
.p { color: #839496 } /* Punctuation */
.cm { color: #586e75; font-style: italic } /* Comment.Multiline */
.cp { color: #586e75; font-style: italic } /* Comment.Preproc */
.c1 { color: #586e75; font-style: italic } /* Comment.Single */
.cs { color: #6c71c4; font-style: italic; } /* Comment.Special */
.gd { color: #839496 } /* Generic.Deleted */
.ge { color: #839496 } /* Generic.Emph */
.gr { color: #839496 } /* Generic.Error */
.gh { color: #839496 } /* Generic.Heading */
.gi { color: #839496 } /* Generic.Inserted */
.go { color: #839496 } /* Generic.Output */
.gp { color: #839496 } /* Generic.Prompt */
.gs { color: #839496 } /* Generic.Strong */
.gu { color: #839496 } /* Generic.Subheading */
.gt { color: #839496 } /* Generic.Traceback */
.kc { color: #859900 } /* Keyword.Constant */
.kd { color: #859900 } /* Keyword.Declaration */
.kn { color: #cb4b16 } /* Keyword.Namespace */
.kp { color: #cb4b16 } /* Keyword.Pseudo */
.kr { color: #859900 } /* Keyword.Reserved */
.kt { color: #859900 } /* Keyword.Type */
.ld { color: #839496 } /* Literal.Date */
.m { color: #2aa198 } /* Literal.Number */
.s { color: #2aa198 } /* Literal.String */
.na { color: #839496 } /* Name.Attribute */
.nb { color: #268bd2 } /* Name.Builtin */
.nc { color: #268bd2 } /* Name.Class */
.no { color: #b58900 } /* Name.Constant */
.nd { color: #cb4b16 } /* Name.Decorator */
.ni { color: #cb4b16 } /* Name.Entity */
.ne { color: #cb4b16 } /* Name.Exception */
.nf { color: #268bd2 } /* Name.Function */
.nl { color: #839496 } /* Name.Label */
.nn { color: #b58900 } /* Name.Namespace */
.nx { color: #839496 } /* Name.Other */
.py { color: #268bd2 } /* Name.Property */
.nt { color: #859900 } /* Name.Tag */
.nv { color: #cd4b16 } /* Name.Variable */
.ow { color: #859900 } /* Operator.Word */
.w { color: #002b36 } /* Text.Whitespace */
.mf { color: #2aa198 } /* Literal.Number.Float */
.mh { color: #2aa198 } /* Literal.Number.Hex */
.mi { color: #2aa198 } /* Literal.Number.Integer */
.mo { color: #2aa198 } /* Literal.Number.Oct */
.sb { color: #2aa198 } /* Literal.String.Backtick */
.sc { color: #2aa198 } /* Literal.String.Char */
.sd { color: #2aa198 } /* Literal.String.Doc */
.s2 { color: #2aa198 } /* Literal.String.Double */
.se { color: #cb4b16 } /* Literal.String.Escape */
.sh { color: #2aa198 } /* Literal.String.Heredoc */
.si { color: #cb4b16 } /* Literal.String.Interpol */
.sx { color: #2aa198 } /* Literal.String.Other */
.sr { color: #2aa198 } /* Literal.String.Regex */
.s1 { color: #2aa198 } /* Literal.String.Single */
.ss { color: #2aa198 } /* Literal.String.Symbol */
.bp { color: #268bd2; } /* font-weight: bold } /* Name.Builtin.Pseudo XXX */
.vc { color: #268bd2 } /* Name.Variable.Class */
.vg { color: #268bd2 } /* Name.Variable.Global */
.vi { color: #268bd2 } /* Name.Variable.Instance */
.il { color: #2aa198 } /* Literal.Number.Integer.Long */
.vm { color: #d33682 } /* dunder object */
.sa { color: #586e75 } /* bytes prefix */
