html {
  overflow-x: hidden; }

/* Under Construction sign */
/* p.construction {
  background-color: orange;
  border-color: black;
  border-style: dashed;
  border-width: .5em 0;
  font-family: monospace;
  font-weight: bold;
  letter-spacing: .1em;
  margin: 1em 0 0 -.5em;
  padding: .25em;
  position: absolute;
  right: -11em;
  text-align: center;
  text-transform: uppercase;
  top: 1em;
  transform: rotate(15deg);
  width: 50em;
  z-index: -9; } */
p.construction {
  background-color: orange;
  border-color: black;
  border-style: dashed;
  border-width: .15em;
  font-family: monospace;
  font-weight: bold;
  height: 1em;
  letter-spacing: .1em;
  margin: 0;
  text-align: center;
  text-transform: uppercase;
  width: 100%; }
p.construction:after {
  content: 'THIS SITE IS UNDER CONSTRUCTION'; }
p.construction span {
  display: none; }
