$def with (ip_address)
<!doctype html>
<html>
<head>
<meta http-equiv=refresh content=10>
<title>Bootstrapping website</title>
<style>
html {
  background-color: #666; }
body {
  background-color: #ccc;
  border-color: #bbb;
  border-style: solid;
  border-width: 0 .25em .25em .25em;
  color: #333;
  margin: 0 auto;
  max-width: 30em;
  padding: 2em 1em; }
h1 {
  font-family: monospace;
  font-size: 4em;
  margin-top: 0;
  text-align: center; }
</style>
</head>
<body>
<h1>Bootstrapping website</h1>
<p>The machine has booted and the core system is being installed. Please wait. This
page will refresh when your system is ready.</p>
</body>
</html>
